//bug crash@close if broken file link error
//Copyright 2016-2017 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#ifdef HAVE_CONFIG
# include <config.h>
#endif
#include "bnDir.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>

#include <albus_log.h>
#ifndef bnDir_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif


#ifdef WIN32
static char *strndup( const char *a, size_t b) {
char *c = (char *)malloc (b+1);
    memset  ( c, 0x00, b+1 );
    strncpy ( c,    a, b   );
    return c;
}
#endif


#include <stdarg.h>
static char *dvstrdup ( const char *format, ... ) {
char Buffer[128000] = "";
va_list arg;
    va_start ( arg, format );
int BufferLen = vsprintf ( Buffer, format, arg );
    va_end   ( arg );
    return strdup ( Buffer );
}//static char *dvstrdup ( const char *format, ... )


void *_bnDir::Show ( void ) {
    D2 ( "this=%p"       , this     );
    D2 ( " vsDir   =%p"  , vsDir    );
    D2 ( " Name    =~%s~", Name     );
    D2 ( " Path    =~%s~", Path     );
    D2 ( " PathName=~%s~", PathName );
    D2 ( " Include =%ld" , Include  );
    D2 ( " Recursiv=%ld" , Recursiv );
    D2 ( " NameBeg =~%s~", NameBeg  );
    D2 ( " NameMid =~%s~", NameMid  );
    D2 ( " NameEnd =~%s~", NameEnd  );
    return NULL;
}//void *_bnDir::Show ( void )


void *_bnDir::Open ( void ) {
    D9 ( "this=%p", this );

    OpenDone = 1;
    if ( Error[0] ) return NULL;
    D9 ( "Path=~%s~ Name=~%s~", Path, Name );

    if ( ! vsDir ) vsDir = new std::vector<_bnDir *>;
    if ( ! vsDir ) SER ( "! vsDir" );
    if ( ! Path  ) SER ( "! Path"  );

    if ( Name ) {
	if ( '*' == Name[0]                     ) NameEnd = strdup  ( Name + 1 );
	if ( '*' == Name[ strlen ( Name ) - 1 ] ) NameBeg = strndup ( Name, strlen ( Name ) - 1 );
    }//if ( Name )
    //Show ();//bug4debug

DIR *Dir = opendir ( Path );
    if ( ! Dir ) SER ( "opendir ( ~%s~ ) = ~%s~", Path, strerror ( errno ) );

dirent *namelist;
    while ( ( namelist = readdir ( Dir ) ) ) {
	if ( strcmp ( namelist->d_name, "."  ) == 0 ) continue;
	if ( strcmp ( namelist->d_name, ".." ) == 0 ) continue;
    long d_type = 0;
#ifndef WIN32
	d_type = namelist->d_type;
#endif
	D9 ( "namelist->d_type=%02d d_name=~%s~", d_type, namelist->d_name );
    int     diNewPush = 1;
    _bnDir *diNew = new _bnDir;
	if ( ! diNew ) SER ( "! diNew" );
	diNew->Name     = strdup  ( namelist->d_name );
	diNew->Path     = strdup  ( Path             );
	diNew->PathName = dvstrdup ( "%s/%s", Path, namelist->d_name );
#ifndef WIN32
	if      ( DT_UNKNOWN == namelist->d_type ) diNew->Include = 0           ;
	else if ( DT_FIFO    == namelist->d_type ) diNew->Include = bnDir_FIFO  ;
	else if ( DT_CHR     == namelist->d_type ) diNew->Include = bnDir_CHR   ;
	else if ( DT_DIR     == namelist->d_type ) diNew->Include = bnDir_DIR   ;
	else if ( DT_BLK     == namelist->d_type ) diNew->Include = bnDir_BLK   ;
	else if ( DT_REG     == namelist->d_type ) diNew->Include = bnDir_FILE  ;
	else if ( DT_SOCK    == namelist->d_type ) diNew->Include = bnDir_SOCK  ;
	else if ( DT_WHT     == namelist->d_type ) diNew->Include = 0           ;
	else if ( DT_LNK     == namelist->d_type ) {
//ios arm64...
#define stat64 stat
	struct stat64 Stat;
	    if ( stat64 ( diNew->PathName, &Stat ) != 0 ) SER ( "stat64 ( ~%s~ ) = ~%s~", diNew->PathName, strerror ( errno ) );
	    if      ( ( Stat.st_mode & S_IFREG  ) == S_IFREG  ) diNew->Include = bnDir_FILELNK;
	    else if ( ( Stat.st_mode & S_IFDIR  ) == S_IFDIR  ) diNew->Include = bnDir_DIRLNK;
	}
#endif
	if ( ! diNew->Include ) SER ( "! diNew->Include" );
	D9 ( "Include=%02d PathName=~%s~", diNew->Include, diNew->PathName );

    //diNew Add
	if ( ! ( Include & diNew->Include ) ) diNewPush = 0;
	if ( ( NameEnd ) && ( strncmp ( diNew->Name + strlen ( diNew->Name ) - strlen ( NameEnd ), NameEnd, strlen ( NameEnd ) ) != 0 ) ) diNewPush = 0;
	if ( ( NameBeg ) && ( strncmp ( diNew->Name                                              , NameBeg, strlen ( NameBeg ) ) != 0 ) ) diNewPush = 0;
	if ( diNewPush ) vsDir->push_back ( diNew );

    //Recursiv
	if ( Recursiv & diNew->Include ) {
	    D1 ( "Recursiv PathName=~%s~", diNew->PathName );
	_bnDir      recDir;
	            recDir.Path     = strdup ( diNew->PathName );
	if ( Name ) recDir.Name     = strdup ( Name            );
		    recDir.Recursiv = Recursiv;
		    recDir.Include  = Include;
		    recDir.vsDir    = vsDir;
		    recDir.Open ();
		    recDir.vsDir    = NULL;
		    if ( recDir.Error[0] ) delete diNew;
		    SIER ( recDir.Error );
	}//if ( Recursiv & diNew->Include )

	if ( ! diNewPush ) delete diNew;

	//break; //bug4debug
    }//while ( ( namelist = readdir ( Dir ) ) )

    Records = vsDir->size();

//free
    //free ( namelist );//2do seg fault
    closedir ( Dir );

    return NULL;

}//char	*_bnDir::Open ( void )


void *_bnDir::Close ( void ) {
    D9 ( "this=%p", this );

    if ( PathName ) free ( PathName ); PathName = NULL;
    if ( Path     ) free ( Path     ); Path     = NULL;
    if ( Name     ) free ( Name     ); Name     = NULL;
    if ( NameBeg  ) free ( NameBeg  ); NameBeg  = NULL;
    if ( NameMid  ) free ( NameMid  ); NameMid  = NULL;
    if ( NameEnd  ) free ( NameEnd  ); NameEnd  = NULL;

    if ( vsDir ) {
	for ( Fokus = 0; Fokus < vsDir->size(); Fokus++ ) {
	_bnDir *clDir = (*vsDir)[Fokus];
	    clDir->Close ();
	    delete clDir; clDir = NULL;
	}
	delete vsDir; vsDir = NULL;
    }//if ( vsDir )

    return NULL;

}//void *_bnDir::Close ( void )


_bnDir *_bnDir::Next ( void ) {
    D9 ( "this=%p", this );

    if ( ! OpenDone ) Open ();
    if ( Error[0]   ) return NULL;
    if ( ! vsDir    ) SER ( "! vsDir" );
    if ( Fokus >= vsDir->size() ) return NULL;

    D9 ( "Fokus=%ld vsDir.size()=%ld", Fokus, vsDir->size() );
    Fokus++;
    return (*vsDir)[Fokus-1];

}//_bnDir *_bnDir::Next ( void )


#ifdef bnDir_MAIN
extern char Error[];


void *bnDirTest ( void ) {
    D9 ( "call" );
    MI ( NULL );
    MI ( NULL );

    //getcwd ( curDir, 1023 );

_bnDir        bnDir;
    //bnDir.Path = strdup ( curDir  );
    //bnDir.Path = strdup ( "DirSave"    );//error=OK
    bnDir.Path = strdup ( "."    );//pwd
    //bnDir.Path = strdup ( "/usr/local/libbncsql/bin" );
    //bnDir.Path = strdup ( "/etc"    );
    //bnDir.Name = strdup ( "*.crt" ); //OK
    //bnDir.Name = strdup ( "*.so"  ); //OK
    //bnDir.Name = strdup ( "libbnc*" ); // OK
    //bnDir.Recursiv  = 0;
    //bnDir.Recursiv += bnDir_DIR;
    //bnDir.Recursiv += bnDir_DIRLNK;
    //bnDir.Include      = 0;
    //bnDir.Include     += bnDir_FILE;
    //bnDir.Include     += bnDir_FILELNK;
    //bnDir.Include     += bnDir_DIR;
    //bnDir.Include     += bnDir_DIRLNK;
    //bnDir.Include     += bnDir_BLK;
    //bnDir.Include     += bnDir_CHR;
    //bnDir.Include     += bnDir_FIFO;
    //bnDir.Include     += bnDir_SOCK;
    while ( _bnDir *diNew = bnDir.Next () ) {
	D2 ( "Include=%02d PathName[%03ld]=~%s~", diNew->Include, bnDir.Fokus, diNew->PathName );
	//diNew->Show ();
	//break;//bug4debug
    }//while ( _bnDir *diNew = bnDir.Next () )
    D9 ( "bnDir.Error=~%s~", bnDir.Error );
    SIER ( bnDir.Error );
    D2 ( "bnDir.Fokus=%ld bnDir.Records=%ld", bnDir.Fokus, bnDir.Records );
    bnDir.Close ();
    MI ( NULL );
    return NULL;

}//void *bnDirTest ( void )


int main ( void ) { bnDirTest (); return 0; }
#endif
