//Copyright 2017-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#include <stdio.h>
#include <string.h>

#include <bncsql.h>
#include <albus_log.h>
#ifndef bncSQLColumn_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif


char *_bncColumn::Show ( void ) {
    D2 ( "this       =%p"    , this       );
    D2 ( " Name      =%10s"  , Name       );
    D2 ( " Type      =0x%08x", Type       );
    D2 ( " *Size     =%8ld"  , *Size      );
    D2 ( " Data      =%p"    , Data       );
    D2 ( " DataExtern=%p"    , DataExtern );
    D2 ( " Scale     =%ld"   , Scale      );
    D2 ( " Format    =%s"    , Format     );
    return NULL;
}//char *_bncColumn::Show ( void )


void *_bncColumn::New ( const void *nData ) {
    D9 ( "this=%p", this );

#ifdef WIN32
    bn_free ( Format );//bug2
#endif
#ifdef __ANDROID__
    bn_free ( Format );//bug2
#endif

    DataExtern = (void *)nData;
    if      ( bn_long & Type ) {
	if ( DataExtern ) Data = DataExtern;
	else              Data = new long;
	if ( ! Format ) {
	    Format = new char[25];
#ifdef WIN32
	    sprintf ( Format, "%%ld"  );//bug2
#else
#ifdef __ANDROID__
	    sprintf ( Format, "%%ld"  );//bug2
#else
	    sprintf ( Format, "%%'ld" );
#endif
#endif
	}
    }
    else if ( bn_long_long & Type ) {
	if ( DataExtern ) Data = DataExtern;
	else              Data = new long long;
	if ( ! Format ) {
	    Format = new char[25];
#ifdef WIN32
	    sprintf ( Format, "%%lld"  );//bug2
#else
#ifdef __ANDROID__
	    sprintf ( Format, "%%lld"  );//bug2
#else
	    sprintf ( Format, "%%'lld" );
#endif
#endif
	}
    }
    else if ( bn_blob & Type ) {
	if ( DataExtern ) Data = DataExtern;
	else              {
	                  Data = (void *)malloc ( sizeof ( unsigned char * ) );
	    memset (      Data, 0x00,             sizeof ( unsigned char * ) );
	}
    }
    else if ( bn_double & Type ) {
	if ( DataExtern ) Data = DataExtern;
	else              Data = new double;
	if ( ! Format ) {
	    Format = new char[25];
#ifdef WIN32
	    sprintf ( Format, "%%0.%ldlf" , Scale );//bug2
#else
#ifdef __ANDROID__
	    sprintf ( Format, "%%0.%ldlf" , Scale );//bug2
#else
	    sprintf ( Format, "%%'0.%ldlf", Scale );
#endif
#endif
	}
    }
    else if ( bn_char & Type ) {
	if ( DataExtern ) Data = DataExtern;
	else              Data = new char[*Size + 1];
    }
    else SER ( "! Type" );

    Clean ();
#ifdef bncSQL_DEBUG
    Show ();
#endif
    return NULL;

}//void *_bncColumn::New ( const void *nData )


void *_bncColumn::Clean ( void ) {
    D9 ( "this=%p", this );

    Type &= ~bn_IS_NULL; //unset NULL
    if ( ! Data ) return NULL;
    if      ( bn_long      & Type ) {      long *longP      =      (long *)Data;            *longP = 0;                }
    else if ( bn_long_long & Type ) { long long *long_longP = (long long *)Data;       *long_longP = 0;                }
    else if ( bn_double    & Type ) {    double *doubleP    =    (double *)Data;          *doubleP = 0;                }
    else if ( bn_char      & Type ) {      char *charP      =      (char *)Data; memset (    charP, 0x00, *Size + 1 ); }
    else if ( bn_blob      & Type ) {
    unsigned char **myData = (unsigned char **)(Data);
	bn_free ( *myData );
	*Size = 0;
    }
    else SER ( "! Type" );

    return NULL;

}//void *_bncColumn::Clean ( void )


void *_bncColumn::Destructor ( void ) {
    D9 ( "this=%p", this );

    D9 ( "Name=%s", Name );
    Clean ();
    if ( Data   == DataExtern   ) Data   = NULL;
    if ( Size   == SizeExtern   ) Size   = NULL;
    if ( Format == FormatExtern ) Format = NULL;

    return NULL;

}//void *_bncColumn::Destructor ( void )


const char *_bncColumn::Get ( void ) {
    D9 ( "this=%p", this );

    if ( bn_long & Type ) {
	memset ( &GetBuffer, 0x00, 100 );
    long *longG = GetLong ();
	if ( Error[0] ) return NULL;
	if ( ! longG  ) return NULL;
	sprintf ( GetBuffer, Format, (long)(*longG) );
	return GetBuffer;
    }
    else if ( bn_long_long & Type ) {
	memset ( &GetBuffer, 0x00, 100 );
    long long *long_longG = GetLongLong ();
	if ( Error[0]     ) return NULL;
	if ( ! long_longG ) return NULL;
	sprintf ( GetBuffer, Format, (long long)(*long_longG) );
	return GetBuffer;
    }
    else if ( bn_double & Type ) {
	memset ( &GetBuffer, 0x00, 100 );
    double *doubleG = GetDouble ();
	if ( Error[0]   ) return NULL;
	if ( ! doubleG  ) return NULL;
	sprintf ( GetBuffer, Format, (double)(*doubleG) );
	return GetBuffer;
    }
    else if ( bn_char & Type ) return GetChar ();
    else SER ( "! Type=0x%x", Type );

    return NULL;

}//const char *_bncColumn::Get ( void )


char *_bncColumn::GetChar ( void ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_char & Type ) ) SER ( "! ( bn_char & Type )" );
    return (char *)Data;
}//char *_bncColumn::GetChar ( void )


long *_bncColumn::GetLong ( void ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_long & Type ) ) SER ( "! ( bn_long & Type )" );
    return (long *)Data;
}//long *_bncColumn::GetLong ( void )


long long *_bncColumn::GetLongLong ( void ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_long_long & Type ) ) SER ( "! ( bn_long_long & Type )" );
    return (long long *)Data;
}


double *_bncColumn::GetDouble ( void ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_double & Type ) ) SER ( "! ( bn_double & Type )" );
    return (double *)Data;
}//double *_bncColumn::GetDouble ( void )


void *_bncColumn::GetBlob ( unsigned char **gData, unsigned long *gLen ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_blob & Type ) ) SER ( "! ( bn_blob & Type )" );
unsigned char **myData = (unsigned char **)(Data);
    *gLen  = *Size;
    *gData = *myData;
    return NULL;
}//void *_bncColumn::GetBlob ( unsigned char **gData, unsigned long *gLen )


void *_bncColumn::Set ( const unsigned char *sData, unsigned long sLen ) {
    D9 ( "this=%p", this );

    if ( ! sData ) {
	Type |= bn_IS_NULL;
	return NULL;
    }

    if ( bn_blob & Type ) SetBlob ( sData, sLen );
    else                  Set     ( sData       );

    return NULL;

}//void *_bncColumn::Set ( const unsigned char *sData, unsigned long sLen )


void *_bncColumn::Set ( const unsigned char *sData ) {
    D9 ( "this=%p", this );
    D9 ( "Type=0x%4x sData=~%s~", Type, sData );

    if ( ! sData ) {
	Type |= bn_IS_NULL;
	return NULL;
    }

    if      ( bn_char   & Type ) SetChar ( sData );
    else if ( bn_long   & Type ) {
    long myLong = 0;
#ifdef WIN32
	sscanf  ( (const char *)sData, "%ld" , &myLong );//bug2
#else
#ifdef __ANDROID__
	sscanf  ( (const char *)sData, "%ld" , &myLong );//bug2
#else
//2do2	sscanf  ( (const char *)sData, "%'ld", &myLong );
	sscanf  ( (const char *)sData, "%ld" , &myLong );//bug2
#endif
#endif
	SetLong ( myLong );
    }
    else if ( bn_long_long & Type ) {
    long long myLongLong = 0;
#ifdef WIN32
	sscanf  ( (const char *)sData, "%lld" , &myLongLong );//bug2
#else
#ifdef __ANDROID__
	sscanf  ( (const char *)sData, "%lld" , &myLongLong );//bug2
#else
//2do2	sscanf  ( (const char *)sData, "%'lld", &myLongLong );
	sscanf  ( (const char *)sData, "%lld" , &myLongLong );//bug2
#endif
#endif
	SetLongLong ( myLongLong );
    }
    else if ( bn_double & Type ) {
    if ( strlen ( (const char *)sData ) > 30 ) SER ( "strlen ( (const char *)sData ) > 30" );
	memset ( &GetBuffer, 0x00, 100 );
	strcpy (  GetBuffer, (const char *)sData );
    double myDouble = 0;
	D9 ( "GetBuffer=~%s~", GetBuffer );
#ifdef WIN32
	sscanf    ( GetBuffer, "%lf" , &myDouble );//bug2
#else
#ifdef __ANDROID__
	sscanf    ( GetBuffer, "%lf" , &myDouble );//bug2
#else
//2do2	sscanf    ( GetBuffer, "%'lf", &myDouble );
	sscanf    ( GetBuffer, "%lf" , &myDouble );//bug2, ok @ ios !!!
#endif
#endif
	D9 ( "myDouble=%f", myDouble );
	SetDouble ( myDouble );
    }
    else if ( bn_blob & Type ) {
	SetBlob ( sData, strlen ( (const char *)sData ) );
    }
    else SER ( "! Type=%ld", Type );

    return NULL;

}//void *_bncColumn::Set ( const unsigned char *sData )


void *_bncColumn::SetChar ( const unsigned char *sData ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_char & Type ) ) SER ( "! ( bn_char & Type )" );
unsigned long sDataLen = strlen ( (const char *)sData );
    if ( sDataLen > (unsigned long)*Size ) SER ( "sDataLen=%ld > *Size=%ld Name=~%s~", sDataLen, *Size, Name );
char *charP = (char *)Data;
    D9 ( "sData=~%s~ sDataLen=%ld *Size=%ld", sData, sDataLen, *Size );
    memcpy ( charP, sData, sDataLen );
    D9 ( "charP=~%s~", charP );
    return NULL;
}//void *_bncColumn::SetChar ( const unsigned char *sData )


void *_bncColumn::SetLong ( long sLong ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_long & Type ) ) SER ( "! ( bn_long & Type )" );
    D9 ( "sLong=%ld", sLong );
long *longP = (long *)Data;
     *longP = sLong;
    return NULL;
}//void *_bncColumn::SetLong ( long sLong )


void *_bncColumn::SetLongLong ( long long sLongLong ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_long_long & Type ) ) SER ( "! ( bn_long_long & Type )" );
    D9 ( "sLong=%lld", sLongLong );
long long *long_longP = (long long *)Data;
	  *long_longP = sLongLong;
    return NULL;
}//void *_bncColumn::SetLongLong ( long long sLongLong )


void *_bncColumn::SetDouble ( double sDouble ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_double & Type ) ) SER ( "! ( bn_double & Type )" );
double *doubleP = (double *)Data;
       *doubleP = sDouble;
    return NULL;
}//void *_bncColumn::SetDouble ( double  sDouble )


void *_bncColumn::SetBlob ( const unsigned char *sData, unsigned long sLen ) {
    D9 ( "this=%p", this );
    if ( ! ( bn_blob & Type ) ) SER ( "! ( bn_blob & Type )" );
    D9 ( "Type=%ld sData=~%p~ sLen=%ld", Type, sData, sLen );

    if ( ! sData ) {
	Type |= bn_IS_NULL;
	return NULL;
    }

    D9 ( "   Data=%p",   Data  );
unsigned char **myData = (unsigned char **)(Data);
    D9 ( " myData=%p",  myData );
    D9 ( "*myData=%p", *myData );

    *Size = sLen;
             *myData = (unsigned char *)new char[ sLen + 1 ];
    memset ( *myData, 0x00 , sLen + 1 );
    memcpy ( *myData, sData, sLen     );

    return NULL;

}//void *_bncColumn::SetBlob ( const unsigned char *sData, unsigned long sLen )
