//Copyright 2017-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#include <bncsql.h>

#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <stdarg.h>

#include <bnSplitPos.h>

//#define bncSQL_DEBUG
#include <albus_log.h>
#ifndef bncSQLTable_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif

#ifdef WIN32
static char *strndup( const char *a, size_t b) {
char *c = (char *)malloc (b+1);
    memset  ( c, 0x00, b+1 );
    strncpy ( c,    a, b   );
    return c;
}
#endif

void *_bncTable::Open ( void ) {
    D8 ( "this=%p", this );

//error check
    if ( Error[0]  ) return NULL;
    if ( Result    ) return NULL;

    if ( ! db      ) db = &dbDefault;

    if ( ( Query ) && ( strncmp ( Query, "bncsql", strlen ( "bncsql" ) ) == 0 ) ) {
	D9 ( "db=%p", db );
	//db->Show ();//bug4debug
    _bncDataBase          *db_new = new _bncDataBase; //2do free
	                   db_new->Driver  = strdup ( "bncsql"    );
	if ( db->Home    ) db_new->Home    = strdup ( db->Home    );
	if ( db->LibPath ) db_new->LibPath = strdup ( db->LibPath );
	db =               db_new;
	//db->Show ();//bug4debug
    }//if ( strncmp ( Query, "bncsql", strlen ( "bncsql" ) ) == 0 )

//open Database
    db->Open (); SIER ( db->Error );

    db->tbSQLBuild ( this ); //modify Query @ Driver
    if ( ! Query ) {
	if ( Name ) SQLQuery ( "select * from %s", Name );
    }
    if (   Error[0] ) return NULL;
    if ( ! Query    ) SER ( "! Query"    );
    if ( ! Query[0] ) SER ( "! Query[0]" );

    D8 ( " Query=~%s~", Query );//best D9

    db->tbOpen(this);	//2do if DriverNotLoaded No Error here !!!
    if ( Error[0] ) return NULL;

//tbColumnInsert raw->Mode
    if ( ! (*Columns).size () ) {
	for ( unsigned long ColumnInsertFokus = 0; ColumnInsertFokus < Fields ; ColumnInsertFokus++ ) {
	_bncColumn *coIns = new _bncColumn;
	    if ( ! coIns ) SER ( "! coIns" );
	    db->tbColumnInsert ( this, coIns, ColumnInsertFokus );
	    if ( Error[0] ) {
		delete coIns;
		return NULL;
	    }
	    ColumnInsert ( coIns );
	    SIER ( coIns->Error );
	    if ( Error[0] ) return NULL;
	}//for ( ColumnInsertFokus = 0; ColumnInsertFokus < Fields ; ColumnInsertFokus++ )
    }//if ( ! Columns.Records )

//number of records check
    if ( ( RecordsLimit >= 1 ) && ( RecordsLimit != Records ) ) SER ( "Wrong Number of Records" );

    return NULL;

}//void *_bncTable::Open ( void )


void *_bncTable::Clean ( void ) {
    D8 ( "this=%p", this );

long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) co->Clean ();

    return NULL;

}//void *_bncTable::Clean ( void )


void *_bncTable::Close ( void ) {
    D8 ( "this=%p", this );

//Close Table via Driver
    if ( db ) {
	 db->tbClose(this);
    }

//clean up
    Clean();

    Error[0]	= 0;
    Records	= 0;
    RecordsLimit= 0;
    Fields	= 0;
    Fokus	= 0;

    if ( Query ) free ( Query );
         Query  = NULL;
    Result	= NULL;

long HaveDataExtern = 0;
long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) {
	if ( co->DataExtern ) {
	    HaveDataExtern = 1;
	    break;
	}
    }

    D8 ( "HaveDataExtern=%ld", HaveDataExtern );
    if ( ! HaveDataExtern ) {
    //remove Columns
    long loop = 0;
	while ( _bncColumn *co = (*this)[loop++] ) delete co;
	(*Columns).clear ();
    }//if ( ! HaveDataExtern )
    return NULL;

}//void *_bncTable::Close ( void )


void *_bncTable::Destructor ( void ) {
    D8 ( "this=%p", this );
    Close ();

long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) delete co;
    (*Columns).clear ();

    return NULL;

}//void *_bncTable::Destructor ( void )


void *_bncTable::Querys ( const char *qQuerys ) {
    D8 ( "this=%p", this );

    D9 ( "qQuerys=[%04ld]~%s~", strlen ( qQuerys ), qQuerys );//D9

_bnSplitPos *bnSplitPos = bnSplitPosNew ( (char *)qQuerys, -1 );
	     bnSplitPos->Quote     = (char *)"\'"; //respect ~;~ in -> values ( 'abc;def' )
	     bnSplitPos->DeQuote   = 0;
	     bnSplitPos->Delimiter = ';';
	     bnSplitPos->Ignore    = (char *)" 	\r\n";
	     bnSplitPos->IgnoreLen = strlen ( bnSplitPos->Ignore );
    while (( bnSplitPos = bnSplitPosNext ( bnSplitPos ) )) {
	Close ();
	//bnSplitPosShow ( bnSplitPos ); //bug4debug
	Query = strndup ( bnSplitPos->Input + bnSplitPos->Beg, bnSplitPos->Len );
	D1 ( "Query=[%04ld]~%s~", strlen ( Query ), Query );//D9

	if ( ! Query                 ) SER ( "! Query" );
	if ( ! Query[0]              ) { continue; }
	if (  7  > strlen ( Query )  ) { continue; }
	if ( '#' ==         Query[0] ) { continue; }
	if ( '-' ==         Query[0] ) { continue; }
	D9 ( "Query=[%04ld]~%s~", strlen ( Query ), Query );//D9

    int HaveRecords = 0;
	while ( Next () ) { Show (); HaveRecords = 1; }
	if ( ! HaveRecords ) {
	    Show ();//Header
	}
	if ( Error[0] ) return NULL;

	//break; //bug4debug
    }//while (( bnSplitPos = bnSplitPosNext ( bnSplitPos ) ))

    return NULL;

}//void *_bncTable::Querys ( const char *QueryS )


void *_bncTable::SQLQuery ( const char *format, ... ) {
    D9 ( "this=%p", this );

    Close ();

    bn_free ( Query );
int SQLQuerySize = 64000;
    Query        = new char[SQLQuerySize];

va_list arg;
    va_start ( arg, format );
int SQLQueryLen = vsprintf ( Query, format, arg );
    va_end   ( arg );
    if ( SQLQueryLen > SQLQuerySize ) SER ( "SQLQueryLen=%d > SQLQuerySize=%d", SQLQueryLen, SQLQuerySize );
    D9 ( "Query=~%s~ Len=%ld", Query, strlen ( Query ) );

    return NULL;

}//void *_bncTable::SQLQuery ( const char *format, ... )


void *_bncTable::tbRecord ( void ) {
    D8 ( "this=%p", this );

    if ( Error[0]		) return NULL;
    if ( ! Records		) return NULL;
    if ( ! Result		) return NULL;
    if ( ! (*Columns).size()	) return NULL;

    if ( Records > 0 ) {//Records count got via cursor
    //Begin of Table
	if ( Fokus < 1 ) {
	    Fokus = 1;
	    return NULL;
	}//if ( Fokus < 1 )
    //End of Table
	if ( Fokus > Records ) {
	    Fokus = Records;
	    return NULL;
	}//if ( Fokus > Records )
    }//if ( Records > 0 )

    D8 ( "this=%p have data !!!", this );

//fetch record
    db->tbRecord(this);
    if ( Error[0] ) return NULL;

    D8 ( "Fokus=%ld Records=%ld", Fokus , Records );
    if ( ! Records ) return NULL;

    if ( ColumnFetchDone ) return this;

    Clean ();
//fetch columns
long loop = 0;
    while ( _bncColumn *myCol = (*this)[loop++] ) {
	D8 ( "myCol=%p", myCol );

    unsigned long  co_len  = 0;
    unsigned char *co_data = NULL;

	db->tbColumn ( this, myCol->Name, &co_data, &co_len );
	if ( Error[0] ) {
	    return NULL;
	}

	setlocale ( LC_ALL, "C" );
	myCol->Set ( co_data, co_len );
	setlocale ( LC_ALL, db->lc_all  );//8.1
	SIER ( myCol->Error );
	if ( Error[0] ) return NULL;

    }//while ( _bncColumn *myCol = (*this)[loop++] )

    return this;

}//void *_bncTable::tbRecord ( void )


void *_bncTable::First ( void ) {
    D8 ( "this=%p", this );
    Open ();
    Fokus = 1;
    return tbRecord();
}//void *_bncTable::First ( void )


void *_bncTable::Next ( void ) {
    D8 ( "this=%p", this );
    Open ();
    Fokus++;
    return tbRecord();
}//void *_bncTable::Next ( void )


void *_bncTable::Prev ( void ) {
    D8 ( "this=%p", this );
    Open ();
    if ( Fokus < 1 ) Fokus = Records;
    else             Fokus--;
    return tbRecord();
}//void *_bncTable::Prev ( void )


void *_bncTable::Last ( void ) {
    D8 ( "this=%p", this );
    Open ();
    Fokus = Records;
    return tbRecord();
}//void *_bncTable::Last ( void )


_bncColumn *_bncTable::ColumnInsert ( const char *ciName, const void *ciData, long ciSize, unsigned long *ciSizeExtern, const char *ciFormat, long ciType ) {
    D8 ( "this=%p", this );

    if ( ! ciName    ) SER ( "! ciName"    );
    if ( ! ciName[0] ) SER ( "! ciName[0]" );
    if ( ! ciType    ) SER ( "! ciType"    );

_bncColumn              *co = new _bncColumn;
                        co->Name   = strdup ( ciName   );
                        co->Type   =          ciType    ;
    if ( ciSizeExtern ) {
                        co->SizeExtern =      ciSizeExtern;
	         free ( co->Size );
                        co->Size       =    co->SizeExtern;
    }//if ( ciSizeExtern )
                       *co->Size   =          ciSize      ;
    if ( ciFormat ) {
                        co->FormatExtern = (char *)ciFormat;
              bn_free ( co->Format );
                        co->Format       = co->FormatExtern;
    }
                        co->New             ( ciData   );
//2do uindex @ iName, only once ...
    (*Columns).push_back ( co );

    return co;

}//_bncColumn *_bncTable::ColumnInsert ( const char *ciName, long ciType, long ciSize, const void *ciData, const char *ciFormat, unsigned long *ciSizeExtern )


_bncColumn *_bncTable::ColumnInsert ( _bncColumn *co ) {
    D8 ( "this=%p", this );

    if ( ! co          )              SER ( "! co"          );
    if ( ! co->Name    ) { delete co; SER ( "! co->Name"    ); }
    if ( ! co->Name[0] ) { delete co; SER ( "! co->Name[0]" ); }
    if ( ! co->Type    ) { delete co; SER ( "! co->Type"    ); }

//2do long loop2 = 0; //uindex @ iName, only once ...
//2do     while ( _bncColumn	*coAP1 = tbAP1->Columns[loop2++] ) {
//2do 	if ( strcmp ( ciName, coAP1->Name ) == 0 ) return (char *)"Error";
//2do     }//while ( _bncColumn	*coAP1 = tbAP1->Columns[loop2++] )

                        co->New ( NULL );
    (*Columns).push_back ( co );

    return co;
}//_bncColumn *_bncTable::ColumnInsert ( _bncColumn *co )


void *_bncTable::ColumnInsert ( _bncTable *tbSrc ) {
    D9 ( "this=%p", this );
long loop = 0;
    while ( _bncColumn *coSrc = (*tbSrc)[loop++] ) {
	this->ColumnInsert ( coSrc->Name, NULL, *coSrc->Size, NULL, NULL, coSrc->Type );
    }
    return NULL;
}//void *_bncTable::ColumnInsert ( _bncTable *tbSrc )


void *_bncTable::Set ( _bncTable *tbSrc ) {
    D9 ( "this=%p", this );
//copy data from tbSrc -> this
long loop = 0;
    while ( _bncColumn *coSrc = (*tbSrc)[loop++] ) {
	if ( _bncColumn *co = (*this)[coSrc->Name] ) {
	    D9 ( "coSrc->Name=~%s~", coSrc->Name );
	    //co->Show ();//bug4debug
	    //coSrc->Show ();//bug4debug
	    co->Clean ();
	    if      ( bn_IS_NULL   & coSrc->Type ) co->Type |= bn_IS_NULL;
	    else if ( bn_char      & coSrc->Type ) { char      *coChar     = coSrc->GetChar     (); co->SetChar     ( (const unsigned char *)coChar     ); }
	    else if ( bn_long      & coSrc->Type ) { long      *coLong     = coSrc->GetLong     (); co->SetLong     (                       *coLong     ); }
	    else if ( bn_long_long & coSrc->Type ) { long long *coLongLong = coSrc->GetLongLong (); co->SetLongLong (                       *coLongLong ); }
	    else if ( bn_double    & coSrc->Type ) { double    *coDouble   = coSrc->GetDouble   (); co->SetDouble   (                       *coDouble   ); }
	    else if ( bn_blob      & coSrc->Type ) {
	    unsigned char *gData = NULL;
	    unsigned long  gLen  =    0;
		coSrc->GetBlob ( &gData, &gLen );
		co->SetBlob    (  gData,  gLen );
	    }
	    else SER ( "unknown coSrc->Type=0x%X", coSrc->Type );
	}//if ( _bncColumn *co = (*this)[coSrc->Name] )
    }//while ( _bncColumn *coSrc = (*tbSrc)[loop++] )
    return NULL;
}//void *_bncTable::Set ( _bncTable *tbSrc )


_bncColumn *_bncTable::operator[] ( const char *coName ) {
    D9 ( "this=%p", this );
long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) {
	if ( strcmp ( co->Name, coName ) == 0 ) return co;
    }
    return NULL;
}//_bncColumn *_bncTable::operator[] ( const char *coName )


_bncColumn *_bncTable::operator[] ( unsigned int coIndex ) {
    D9 ( "this=%p", this );
    D9 ( "coIndex=%ld (*Columns).size()=%ld", coIndex, (*Columns).size() );
    if ( coIndex >= (*Columns).size () ) return NULL;
    return (*Columns)[ coIndex ];
}//_bncColumn *_bncTable::operator[] ( unsigned int coIndex )


const char *_bncTable::Get ( const char *fName ) {
    D9 ( "this=%p", this );
    if ( _bncColumn *co = (*this)[fName] ) return co->Get ();
    return NULL;
}//const char *_bncTable::Get ( const char *fName )


char *_bncTable::GetChar ( const char *fName ) {
    D9 ( "this=%p", this );
    if ( _bncColumn *co = (*this)[fName] ) return co->GetChar ();
    return NULL;
}//char *_bncTable::GetChar ( const char *fName )


long *_bncTable::GetLong ( const char *fName ) {
    D9 ( "this=%p", this );
    if ( _bncColumn *co = (*this)[fName] ) return co->GetLong ();
    return NULL;
}//long *_bncTable::GetLong ( const char *fName )


double *_bncTable::GetDouble ( const char *fName ) {
    D9 ( "this=%p", this );
    if ( _bncColumn *co = (*this)[fName] ) return co->GetDouble ();
    return NULL;
}//double *_bncTable::GetDouble ( const char *fName )
