//Copyright 2017-2018 blacknoise.at
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
//
#include <bncsql.h>

#include <stdio.h>
#include <locale.h>
#include <sstream>

#include <albus_log.h>
#ifndef bncSQL_DEBUG
# undef  MI
# define MI
# undef  D8
# define D8
# undef  D9
# define D9
#endif


std::vector<std::string> _bncTable::bnSplit ( const std::string &sSplit, char delim ) {
    D9 ( "this=%p", this );
std::vector<std::string> items;
std::string item;
std::stringstream ss ( sSplit );
    while ( getline ( ss, item, delim ) ) items.push_back ( item );

    return items;

}//std::vector<std::string> _bncTable::bnSplit ( const std::string &sSplit, char delim )


void _bncTable::bnReplace ( std::string &str, const std::string &strOut, const std::string& str_In ) {
    if ( ! strOut.length() ) return;
int pos = 0;
    while ( ( pos = str.find ( strOut, pos ) ) != (int)std::string::npos ) {
	str.replace ( pos, strOut.length(), str_In );
	pos += str_In.length ();
    }
}//void _bncTable::Replace ( std::string &str, const std::string &strOut, const std::string& str_In )


void *_bncTable::TransactionBegin ( void ) {
    D9 ( "this=%p", this );

    if ( Error[0] ) return NULL;
//begin transaction
    SQLQuery ( "begin;" );
    Open  (); SIER ( Error );
    Close ();

    return NULL;

}//void *_bncTable::TransactionBegin ( void )


void *_bncTable::TransactionCommit ( void ) {
    D9 ( "this=%p", this );

    if ( Error[0] ) return NULL;
//commit transaction
    SQLQuery ( "commit;" );
    Open  (); SIER ( Error );
    Close ();

    return NULL;

}//void *_bncTable::TransactionCommit ( void )


void *_bncTable::TransactionRollback ( void ) {
    D9 ( "this=%p", this );

    if ( Error[0] ) return NULL;
//rollback transaction
    SQLQuery ( "rollback;" );
    Open (); SIER ( Error );
    Close ();

    return NULL;

}//void *_bncTable::TransactionRollback ( void )


void *_bncTable::TransactionEnd ( char *trError ) {
    D9 ( "this=%p", this );

    if ( trError[0] ) TransactionRollback ();
    else              TransactionCommit   ();

    return trError;

}//void *_bncTable::TransactionEnd ( char *trError )


void *FormatField ( std::string &sFormat, _bncTable *tb, _bncColumn *co ) {
    D9 ( "tb=%p", tb );
    D9 ( "co=%p", co );
    if ( ' ' != tb->db->Quote ) { sFormat += tb->db->Quote; }
                                  sFormat += co->Name;
    if ( ' ' != tb->db->Quote ) { sFormat += tb->db->Quote; }

    return NULL;

}//void *FormatField ( std::string &sFormat, _bncTable *tb, _bncColumn *co )


void *FormatValue ( std::string &fValue, _bncTable *tb, _bncColumn *co ) {
    D9 ( "tb=%p", tb );
    D9 ( "co=%p", co );

char fValueBuffer[100];
    if      ( bn_IS_NULL & co->Type ) {
	fValue += "NULL";
    }
    else if ( bn_char    & co->Type ) {
	fValue += co->Get();
	tb->bnReplace ( fValue, "'", "\\'" );
	fValue.insert ( 0, 1, '\'' );
	fValue             += '\''  ;
    }
    else if ( bn_long    & co->Type ) {
	memset  ( &fValueBuffer, 0x00, 100 );
	sprintf (  fValueBuffer, "%ld", *co->GetLong()  );
	fValue +=  fValueBuffer;
    }
    else if ( bn_long_long & co->Type ) {
	memset  ( &fValueBuffer, 0x00, 100 );
	sprintf (  fValueBuffer, "%lld", *co->GetLongLong()  );//bug2
	fValue +=  fValueBuffer;
    }
    else if ( bn_double  & co->Type ) {
	memset  ( &fValueBuffer, 0x00, 100 );
	sprintf (  fValueBuffer, "%f", *co->GetDouble() );
	fValue +=  fValueBuffer;
    }
    else if ( bn_blob & co->Type ) {
    unsigned long  blob_len;
    unsigned char *blob_data;
	co->GetBlob ( &blob_data, &blob_len );
	if ( ! blob_len ) {
	    fValue += "NULL";
	    return NULL;
	}
	fValue += "0x";
    char *xxxa= (char *)blob_data;
	for ( unsigned long xx=0 ; xx < blob_len ; xx++ ) {
	    memset  ( &fValueBuffer, 0x00, 100 );
	    sprintf (  fValueBuffer, "%02X", (unsigned char)xxxa[xx] );
	    fValue +=  fValueBuffer;
	}
    }

    return NULL;

}//void *FormatValue ( std::string &sFormat, _bncTable *tb, _bncColumn *co )


void *_bncTable::Insert ( void ) {
    D8 ( "this=%p", this );

    if ( Error[0]         ) return NULL;
    if ( ! Name           ) SER ( "! Name"             );
    if ( ! Name[0]        ) SER ( "! Name[0]"          );
    if ( ! (*Columns).size() ) SER ( "! (*Columns).size()"   );
    if (   Query          ) SER ( "Query"              );
    if (   Result         ) SER ( "Table already open" );
    if (   Records        ) SER ( "Table already open" );
    if ( ! db             ) db = &dbDefault;
//open Database
    db->Open (); SIER ( db->Error );

//check @ driver 4 SQLBuild
    db->tbSQLBuild(this);
    if ( Error[0] ) return NULL;

std::string Fields;
std::string Values;
    setlocale ( LC_ALL, (char *)"C" );
long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) {
	//co->Show ();
	co->Type &= ~bn_IS_NULL; //unset NULL

	if ( bn_AUTO_INCREMENT & co->Type ) {
	    if      ( bn_long      & co->Type )
		if ( ! *co->GetLong     () ) continue;
	    else if ( bn_long_long & co->Type )
		if ( ! *co->GetLongLong () ) continue;
	}

    //Fields
    std::string sFormatField;
	FormatField ( sFormatField, this, co );
	D9 ( "sFormatField=~%s~", sFormatField.c_str () );
	if ( Fields.size() ) Fields += ",";
	     Fields += sFormatField;

    //Values
    std::string sFormatValue;
	FormatValue ( sFormatValue, this, co );
	D9 ( "sFormatValue=~%s~", sFormatValue.c_str () );

	if ( Values.size() ) Values += ",";
	     Values += sFormatValue;

    }//while ( _bncColumn *co = (*this)[loop++] )
    setlocale ( LC_ALL, db->lc_all  );
    D8 ( "Fields.size()=%ld Fields.c_str()=~%s~", Fields.size(), Fields.c_str() );
    D8 ( "Values.size()=%ld Values.c_str()=~%s~", Values.size(), Values.c_str() );
    Query = new char[ 50 + Fields.size() + Values.size() + strlen ( Name ) ];
    sprintf ( Query, "insert into %c%s%c (%s) values (%s);", db->Quote, Name, db->Quote, Fields.c_str(), Values.c_str() );
    D8 ( "Query=~%s~", Query );

//Action
    First (); SIER ( Error );

    return NULL;

}//void *_bncTable::Insert ( void )


void *_bncTable::Update ( void ) {
    D9 ( "this=%p", this );

    if ( Error[0]         ) return NULL;
    if ( ! Name           ) SER ( "! Name"             );
    if ( ! Name[0]        ) SER ( "! Name[0]"          );
    if ( ! (*Columns).size() ) SER ( "! (*Columns).size()"   );
    if ( ! db             ) db = &dbDefault;
//open Database
    db->Open (); SIER ( db->Error );

//check @ driver 4 SQLBuild
    db->tbSQLBuild(this);
    if ( Error[0] ) return NULL;

std::string Where;
std::string Set;
    setlocale ( LC_ALL, (char *)"C" );
long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) {
	//co->Show ();
	co->Type &= ~bn_IS_NULL; //unset NULL
    std::string sFormatField;
	FormatField ( sFormatField, this, co );

    std::string sFormatValue;
	FormatValue ( sFormatValue, this, co );

	if ( bn_PRI_KEY & co->Type ) {
	    Where += sFormatField;
	    Where += "=";
	    Where += sFormatValue;
	}
	else {
	    if ( ! sFormatValue.size() ) continue;
	    if ( Set.size() ) Set += ",";
	    Set += sFormatField;
	    Set += "=";
	    Set += sFormatValue;
	}
    }//while ( _bncColumn *co = (*this)[loop++] )
    setlocale ( LC_ALL, db->lc_all  );

    if ( ! Where.size() ) SER ( "! Where.size()" );
    if ( ! Set.size()   ) SER ( "! Set.size()"   );
    D9 ( "Where.size()=%ld Where.c_str()=~%s~", Where.size(), Where.c_str() );
    D9 ( "Set.size()=%ld Set.c_str()=~%s~", Set.size(), Set.c_str() );

_bncTable      tbUD;
	      tbUD.Query = new char[ 50 + Where.size() + Set.size() + strlen ( Name ) ];
    sprintf ( tbUD.Query, "update %c%s%c set %s where %s", db->Quote, Name, db->Quote, Set.c_str(), Where.c_str() );
    D9 ( "Query=~%s~", tbUD.Query );
//Action
    tbUD.First (); SIER ( tbUD.Error );

    return NULL;

}//void *_bncTable::Update ( void )


void *_bncTable::Delete ( void ) {
    D8 ( "this=%p", this );

    if ( Error[0]         ) return NULL;
    if ( ! Name           ) SER ( "! Name"             );
    if ( ! Name[0]        ) SER ( "! Name[0]"          );
    if ( ! (*Columns).size() ) SER ( "! (*Columns).size()"   );
    if ( ! db             ) db = &dbDefault;
//open Database
    db->Open (); SIER ( db->Error );

//check @ driver 4 SQLBuild
    db->tbSQLBuild(this);
    if ( Error[0] ) return NULL;

std::string Where;
    setlocale ( LC_ALL, (char *)"C" );
long loop = 0;
    while ( _bncColumn *co = (*this)[loop++] ) {
	//co->Show ();
	if ( bn_PRI_KEY & co->Type ) {
	std::string sFormatField;
	    FormatField ( sFormatField, this, co );
	std::string sFormatValue;
	    FormatValue ( sFormatValue, this, co );
	    Where += sFormatField;
	    Where += "=";
	    Where += sFormatValue;
	    break;
	}
    }//while ( _bncColumn *co = (*this)[loop++] )
    setlocale ( LC_ALL, db->lc_all  );

    if ( ! Where.size() ) SER ( "! Where.size()" );
    D8 ( "Where.size()=%ld Where.c_str()=~%s~", Where.size(), Where.c_str() );

_bncTable     tbUD;
	      tbUD.db = db;
	      tbUD.Query = new char[ 50 + Where.size() + strlen ( Name ) ];
    sprintf ( tbUD.Query, "delete from %c%s%c where %s", db->Quote, Name, db->Quote, Where.c_str() );
    D8 ( "Query=~%s~", tbUD.Query );

//Action
    tbUD.First (); SIER ( tbUD.Error );

    return NULL;

}//void *_bncTable::Delete ( void )


void *bncSQLBuildWhere ( std::string &bwDST, _bncTable *tb, const char *bwWhere ) {
    D9 ( "tb=%p", tb );

    if ( ! tb ) return NULL;

char *Error = tb->Error;
    if ( ! bwWhere ) SER ( "! bwWhere" );
    bwDST = "";

    tb->db->Open (); SIER ( tb->db->Error );//4 tb->db->Quote

std::vector<std::string> vsWhere = tb->bnSplit ( bwWhere, ',' );
    setlocale ( LC_ALL, (char *)"C" );
    for ( unsigned long vsWherePos = 0; vsWherePos < vsWhere.size(); vsWherePos++ ) {
    std::string sWhere = vsWhere[vsWherePos];
	D9 ( "[%03ld/%03ld]Where=~%s~", vsWherePos + 1, vsWhere.size(), sWhere.c_str() );

    _bncColumn *co = (*tb)[ sWhere.c_str() ];
	if ( ! co ) {
	    setlocale ( LC_ALL, tb->db->lc_all );
	    SER ( "! co [%s]", sWhere.c_str() );
	}
	//co->Show ();

    std::string sFormatField;
	FormatField ( sFormatField, tb, co );

    std::string sFormatValue;
	FormatValue ( sFormatValue, tb, co );

	if ( ! sFormatValue.size() ) {
	    setlocale ( LC_ALL, tb->db->lc_all );
	    SER ( "! sFormatValue.size()" );
	}
	if ( bwDST.size() ) bwDST += " and ";//2do if '%' == sFormatValue[x] -> like
	     bwDST += sFormatField;
	     bwDST += "=";
	     bwDST += sFormatValue;

    }//for ( unsigned long vsWherePos = 0; vsWherePos < vsWhere.size(); vsWherePos++ )
    setlocale ( LC_ALL, tb->db->lc_all );
    D8 ( "bwDST.c_str()=~%s~", bwDST.c_str() );

    return NULL;

}//void *bncSQLBuildWhere ( std::string &bwDST, _bncTable *bwTB, const char *bwWhere )


void *bncSQLBuildSet ( std::string &bsDST, _bncTable *tb, const char *bsSet ) {
    D9 ( "tb=%p", tb );

    if ( ! tb ) return NULL;

char *Error = tb->Error;
    if ( ! bsSet ) SER ( "! bsSet" );
    bsDST = "";

    tb->db->Open (); SIER ( tb->db->Error );//4 tb->db->Quote

std::vector<std::string> vsSet = tb->bnSplit ( bsSet, ',' );
    setlocale ( LC_ALL, (char *)"C" );
    for ( unsigned long vsSetPos = 0; vsSetPos < vsSet.size(); vsSetPos++ ) {
    std::string sSet = vsSet[vsSetPos];
	D9 ( "[%03ld/%03ld]Set=~%s~", vsSetPos + 1, vsSet.size(), sSet.c_str() );

    _bncColumn *co = (*tb)[ sSet.c_str() ];
	if ( ! co ) {
	    setlocale ( LC_ALL, tb->db->lc_all );
	    SER ( "! co [%s]", sSet.c_str() );
	}
	//co->Show ();

    std::string sFormatField;
	FormatField ( sFormatField, tb, co );

    std::string sFormatValue;
	FormatValue ( sFormatValue, tb, co );

	if ( ! sFormatValue.size() ) {
	    setlocale ( LC_ALL, tb->db->lc_all );
	    SER ( "! sFormatValue.size()" );
	}
	if ( bsDST.size() ) bsDST += ",";
	     bsDST += sFormatField;
	     bsDST += "=";
	     bsDST += sFormatValue;

    }//for ( unsigned long vsSetPos = 0; vsSetPos < vsSet.size(); vsSetPos++ )
    setlocale ( LC_ALL, tb->db->lc_all );
    D9 ( "bsDST.c_str()=~%s~", bsDST.c_str() );

    return NULL;

}//void *bncSQLBuildSet ( std::string &bsDST, _bncTable *tb, const char *bsSet )


//2do char	*bncSQLBuildOrder	( char *boDST, _bncTable *boTB, char *boOrder ) {
//2do PFLID
//2do _bncTable	*tb = boTB;//4 cp 2 driver
//2do     if ( ! tb ) return NULL;
//2do char	*Error = tb->Error;
//2do     if ( ! boDST	) SER ( "! boDST"	);
//2do     if ( ! boOrder	) SER ( "! boOrder"	);
//2do     tb->db->Open ();	SIER ( tb->db->Error );//4 tb->db->Quote
//2do _bnStrSplit  spWhere;
//2do     spWhere.In = boOrder;
//2do     spWhere.sCol = ',';	//seperator Column
//2do     while ( char *myName = spWhere.Next () ) {
//2do     _bncColumn	*myCol = tb->Columns[ myName ];
//2do 	if ( ! myCol ) SER2 ( "! myCol [%s]", myName );
//2do     //Name
//2do 	if ( boDST[0] ) strcat ( boDST, ", " );
//2do 	sprintf ( boDST, "%s%c%s%c", boDST, tb->db->Quote,  myCol->Name, tb->db->Quote	);
//2do     }//while ( char *myName = bnStrSplit.Next () )
//2do     return Error;
//2do }//char	*bncSQLBuildOrder	( char *boDST, _bncTable *boTB, char *boOrder )


void *_bncTable::InsertUpdate ( const char *iuWhere, const char *iuSet ) {
    D8 ( "this=%p", this );

    if ( Error[0]         ) return NULL;
    if ( ! Name           ) SER ( "! Name"             );
    if ( ! Name[0]        ) SER ( "! Name[0]"          );
    if ( ! (*Columns).size() ) SER ( "! (*Columns).size()"   );
    if ( ! iuWhere        ) SER ( "! iuWhere"          );
    if ( ! iuSet          ) SER ( "! iuUpdate"         );
    if ( ! db             ) db = &dbDefault;
//open Database
    db->Open (); SIER ( db->Error );

//check @ driver 4 SQLBuild
    db->tbSQLBuild(this); if ( Error[0] ) return NULL;

std::string Where;
    bncSQLBuildWhere ( Where, this, iuWhere ); SIER ( Error );
    if ( ! Where.size() ) SER ( "! Where.size()" );
    D8 ( "Where.size()=%ld Where.c_str()=~%s~", Where.size(), Where.c_str() );

_bncTable      tbIU;
	      tbIU.db = this->db;
//strcpy ( tbIU.Name, iuTB->Name );
	      tbIU.Query = new char[ 50 + Where.size() + strlen ( Name ) ];
    sprintf ( tbIU.Query, "select * from %c%s%c where %s", db->Quote, Name, db->Quote, Where.c_str() );
    D8 ( "Query=~%s~", tbIU.Query );
    tbIU.First (); SIER ( tbIU.Error );

//Insert/Update
    if ( ! tbIU.Records ) {
	D8 ( "Insert !" );
	tbIU.Close ();
	Insert (); SIER ( Error );
    }//if ( ! tbIU.Records )
    else {
	D8( "Update !" );
	tbIU.Close ();

    std::string Set;
	bncSQLBuildSet ( Set, this, iuSet ); SIER ( Error );
	if ( ! Set.size() ) SER ( "! Set.size()" );
	D8 ( "Where.size()=%ld Where.c_str()=~%s~", Where.size(), Where.c_str() );
	D8 ( "Set.size()=%ld Set.c_str()=~%s~", Set.size(), Set.c_str() );
		  tbIU.Query = new char[ 50 + Where.size() + Set.size() + strlen ( Name ) ];
	sprintf ( tbIU.Query, "update %c%s%c set %s where %s", db->Quote, Name, db->Quote, Set.c_str(), Where.c_str() );
	D8 ( "Query=~%s~", tbIU.Query );
//Action
	tbIU.First (); SIER ( tbIU.Error );
    }//else if ( ! tbIU.Records )
//Insert/Update done

    return NULL;

}//void *_bncTable::InsertUpdate ( const char *iuWhere, const char *iuSet )


void *_bncTable::CreateInclude ( _bncTable *tbSrc ) {
    D8 ( "this=%p", this );

std::string ciHeader;
std::string ciConstructor;
std::string ciDestructor;

char Buffer100[1000];
    memset  ( Buffer100, 0x00, 1000 );
    sprintf ( Buffer100, "%13s *%s;\n", "_bncTable", "tb" );
    ciHeader += Buffer100;

long loop = 0;
    while ( _bncColumn *co = (*tbSrc)[loop++] ) {
	//co->Show (); //bug4debug

    char coName[100] = "";
	sprintf ( coName, "\"%s\"", co->Name );

    char coData[100] = "";
	sprintf ( coData, "&%s"   , co->Name );

    char coFormat[100] = "";
	if ( co->FormatExtern ) sprintf ( coFormat, "\"%s\"", co->FormatExtern );
	else                    sprintf ( coFormat, "NULL"                     );

    char coTypeAdd[100] = "";
	if ( co->Type & bn_AUTO_INCREMENT ) strcat ( coTypeAdd, " + bn_AUTO_INCREMENT" );
	if ( co->Type & bn_PRI_KEY        ) strcat ( coTypeAdd, " + bn_PRI_KEY"        );

    char coSizeExtern[100] = "";
	sprintf ( coSizeExtern, "NULL" );

	if      ( bn_char      & co->Type ) {
	    sprintf ( Buffer100, "%13s  %s[%ld];\n", "char", co->Name, *co->Size + 1 );
	    ciHeader += Buffer100;
	    sprintf ( Buffer100, "        tb->ColumnInsert (%20s,%20s,%5ld,%20s,%8s, bn_char     %s );\n", coName, coData, *co->Size, coSizeExtern, coFormat, coTypeAdd );
	    ciConstructor += Buffer100;
	}//if      ( bn_char      & co->Type )
	else if ( bn_long      & co->Type ) {
	    sprintf ( Buffer100, "%13s  %s;\n", "long", co->Name );
	    ciHeader += Buffer100;
	    sprintf ( Buffer100, "        tb->ColumnInsert (%20s,%20s,%5ld,%20s,%8s, bn_long     %s );\n", coName, coData,   (long)0, coSizeExtern, coFormat, coTypeAdd );
	    ciConstructor += Buffer100;
	}//else if ( bn_long      & co->Type )
	else if ( bn_long_long & co->Type ) {
	    sprintf ( Buffer100, "%13s  %s;\n", "long long", co->Name );
	    ciHeader += Buffer100;
	    sprintf ( Buffer100, "        tb->ColumnInsert (%20s,%20s,%5ld,%20s,%8s, bn_long_long%s );\n", coName, coData,   (long)0, coSizeExtern, coFormat, coTypeAdd );
	    ciConstructor += Buffer100;
	}//else if ( bn_long_long & co->Type )
	else if ( bn_double    & co->Type ) {
	    sprintf ( Buffer100, "%13s  %s;\n", "double", co->Name );
	    ciHeader += Buffer100;
	    sprintf ( Buffer100, "        tb->ColumnInsert (%20s,%20s,%5ld,%20s,%8s, bn_double   %s );\n", coName, coData,   (long)0, coSizeExtern, coFormat, coTypeAdd );
	    ciConstructor += Buffer100;
	}//else if ( bn_long_long & co->Type )
	else if ( bn_blob      & co->Type ) {
	    sprintf ( Buffer100, "%13s *%s;\n"   , "unsigned char", co->Name );
	    ciHeader += Buffer100;
	    sprintf ( Buffer100, "%13s  %sLen;\n", "unsigned long", co->Name );
	    ciHeader += Buffer100;
	    sprintf ( coData      , "%s"    , co->Name );
	    sprintf ( coSizeExtern, "&%sLen", co->Name );
	    sprintf ( Buffer100, "        tb->ColumnInsert (%20s,%20s,%5ld,%20s,%8s, bn_blob     %s );\n", coName, coData,   (long)0, coSizeExtern, coFormat, coTypeAdd );
	    ciConstructor += Buffer100;
	    sprintf ( Buffer100, "        bn_free (%20s );", co->Name );
	    ciDestructor += Buffer100;
	}//else if ( bn_blob      & co->Type )
	else SER ( "unknown co->Type=0x%x co->Name=~%s~", co->Type, co->Name );

    }//while ( _bncColumn *co = (*tbSrc)[loop++] )

    SQLQuery ( "\
#ifndef %s_H\n\
#define %s_H\n\
\n\
#include <bncsql.h>\n\
\n\
\n\
class _%s {\n\
public:\n\
%s\n\
    _%s() {\n\
        memset ( this, 0x00, sizeof ( _%s ) );\n\
        tb       = new _bncTable;\n\
        tb->Name = strdup ( \"%s\" );\n\
%s\
    }\n\
\n\
    ~_%s() {\n\
%s\n\
        delete tb;\n\
    }\n\
\n\
};\n\
#endif\n\
", Name, Name, Name, ciHeader.c_str(), Name, Name, Name, ciConstructor.c_str(), Name, ciDestructor.c_str() );

    return NULL;

}//void *_bncTable::CreateInclude ( _bncTable *tbSrc )


void *_bncTable::CreateTable ( char ctMode, _bncTable *tbSrc ) {
    D9 ( "this=%p", this );
//2do index, defaults
    if ( ! tbSrc    ) tbSrc = this;
    if (   Error[0] ) return NULL;
    if ( ! Name     ) SER ( "! Name"    );
    if ( ! Name[0]  ) SER ( "! Name[0]" );
    if (   Query    ) SER ( "Query"     );
    if ( ! db       ) db = tbSrc->db;
    if ( ! db       ) SER ( "! db"      );
    if ( ! (*tbSrc->Columns).size() ) SER ( "! (*tbSrc->Columns).size()" );

//create [temporary] table
    if ( 1 == ctMode ) Query = strdup ( "create table"           );
    else               Query = strdup ( "create temporary table" );
    db->Open ();
//check @ driver 4 SQLBuild
    db->tbSQLBuild(this); if ( Error[0] ) return NULL;
    if ( ! CreateTableOptions ) SER ( "! CreateTableOptions" );
    D9 ( "CreateTableOptions=%s", CreateTableOptions );

std::string ctHeader;
    ctHeader = Query;
    bn_free ( Query );
//create [temporary] table done

//Columns
std::string ctColumns;
long loop = 0;
    while ( _bncColumn *co = (*tbSrc)[loop++] ) {
	co->Show ();
    char *coDEF = db->coDEF ( co ); //SIER ( db->Error );
	if ( ! coDEF    ) SER ( "! coDEF"    );
	if ( ! coDEF[0] ) SER ( "! coDEF[0]" );
	D9 ( "coDEF=~%s~", coDEF );
	if ( ctColumns.size() ) ctColumns += ",";
	     ctColumns += "\n";
	     ctColumns += coDEF;
    }//while ( _bncColumn *co = (*tbSrc)[loop++] )
    if ( ! ctColumns.size() ) SER ( "! ctColumns.size()" );
    D9 ( "ctColumns.size()=%ld ctColumns.c_str()=~%s~", ctColumns.size(), ctColumns.c_str() );
//Columns done

    SQLQuery ( "%s %c%s%c (%s\n) %s;\n", ctHeader.c_str(), db->Quote, Name, db->Quote, ctColumns.c_str(), CreateTableOptions );
    D9 ( "Query=~%s~", Query );

    return NULL;

}//void *_bncTable::CreateTable ( char ctMode, _bncTable *tbSrc )
